import logging
import os

from jinja2 import Template

from .cli import task
from . import portinus, restart, systemd, monitor

_script_dir = os.path.dirname(os.path.realpath(__file__))
template_dir = os.path.join(_script_dir, 'templates')
service_dir = '/usr/local/portinus-services'


def get_instance_dir(name):
    return os.path.join(service_dir, name)


def get_template(file_name):
    template_file = os.path.join(template_dir, file_name)
    with open(template_file) as f:
        template_contents = f.read()

    return Template(template_contents)


class Application(object):

    log = logging.getLogger()

    def __init__(self, name, source=None, environment_file=None, restart_schedule=None):
        self.name = name
        self._environment_file = portinus.EnvironmentFile(name, environment_file)
        self._service = portinus.Service(name, source)
        self._restart_timer = restart.Timer(name, restart_schedule=restart_schedule)
        self._monitor_service = monitor.Service(name)

    def _create_service_dir(self):
        try:
            os.mkdir(service_dir)
        except FileExistsError:
            pass

    def exists(self):
        return self._service.exists()

    def ensure(self):
        self._create_service_dir()
        self._environment_file.ensure()
        self._service.ensure()
        self._restart_timer.ensure()
        self._monitor_service.ensure()

    def remove(self):
        self._service.remove()
        self._environment_file.remove()
        self._restart_timer.remove()
        self._monitor_service.remove()



import docker
import logging
import os
import subprocess
import sys

import portinus

log = logging.getLogger(__name__)


def check_permissions():
    if os.geteuid() != 0:
        log.error("This script must be run as root!")
        raise PermissionError("You must be root")


def run(name):
    check_permissions()

    systemd_service = portinus.systemd.Unit(name)
    monitored_compose_containers = get_monitored_compose_containers(name)

    for container in monitored_compose_containers:
        log.debug("Checking container {container_id}, name: '{name}'".format(container_id=container.id, name=container.attrs['Name']))
        if not check_container_health(container):
            log.info("Container {container_id}, name: '{container_name}'. Restarting stack for {name}".format(container_id=container.id, container_name=container.attrs['Name'], stack_name=name))
            print("Container '{container_name} found unhealthy. Restarting the {name} stack...".format(container_name=container.attrs['Name'], name=name))
            systemd_service.restart()
            return False
    print("No unhealthy containers found")


def check_container_health(container):
    status = container.attrs["State"]["Health"]["Status"]
    return status != "unhealthy"


def get_monitored_compose_containers(name):
    compose_container_ids = get_compose_container_ids(name)
    monitored_containers = get_monitored_containers()
    monitored_compose_containers = []

    for container_id in compose_container_ids:
        valid_container = next((x for x in monitored_containers if x.id == container_id), None)
        if valid_container: monitored_compose_containers.append(valid_container)

    log.debug("Found {count} monitored containers from docker-compose".format(count=len(monitored_compose_containers)))
    return monitored_compose_containers


def get_compose_container_ids(name):
    compose_source = portinus.portinus.ComposeSource(name, None)
    service_script = compose_source.service_script

    compose_output = subprocess.check_output([service_script, "ps", "-q"], stderr=subprocess.DEVNULL).decode("utf-8")
    container_list = compose_output.split("\n")

    filtered_container_list = list((x for x in container_list if x))

    log.debug("Found {count} total containers from docker-compose".format(count=len(filtered_container_list)))
    return filtered_container_list


def get_monitored_containers():
    client = docker.from_env()

    all_containers = client.containers.list()

    monitored_containers = []

    for container in all_containers:
        if 'Health' in container.attrs['State']:
            monitored_containers.append(container)

    log.debug("Found {count} monitored containers from docker".format(count=len(monitored_containers)))
    return monitored_containers



import logging
import os
import subprocess

log = logging.getLogger(__name__)


class Unit(object):

    def __init__(self, name, type="service", content=None):
        self.name = name
        self.service_name = "portinus-{name}.{type}".format(name=name, type=type)
        self.service_file_path = os.path.join("/etc/systemd/system", self.service_name)
        self._content = content

        try:
            subprocess.check_output(["systemctl", "--help"])
        except FileNotFoundError as e:
            log.error("Unable to find systemctl!")
            raise(e)
        log.debug("Initialized systemd.Unit for '{name}' with type '{type}'".format(name=name, type=type))

    def set_content(self, content):
        self._content = content

    def create_service_file(self):
        log.info("Creating/updating service file for '{name}' at '{service_file_path}'".format(name=self.name, service_file_path=self.service_file_path))
        with open(self.service_file_path, "w") as f:
            f.write(self._content)

    def ensure(self, restart=True, enable=True, content=None):
        if content: self.set_content(content)
        self.create_service_file()
        self.reload()
        if restart: self.restart()
        if enable: self.enable()

    def remove(self):
        try:
            if os.path.exists(self.service_file_path):
                self.stop()
                self.disable()
        except subprocess.CalledProcessError:
            pass

        log.info("Removing service file for {name} from {service_file_path}".format(name=self.name, service_file_path=self.service_file_path))
        try:
            os.remove(self.service_file_path)
            log.debug("Successfully removed service file")
        except FileNotFoundError:
            log.debug("No service file found")
        self.reload()

    def _systemctl(self, args, stderr=None):
        try:
            subprocess.check_output(["systemctl"] + args, stderr=stderr)
        except subprocess.CalledProcessError as e:
            log.error("Failed to run systemctl with parameters {args}".format(args=args))
            raise(e)

    def reload(self):
        log.info("Reloading daemon files")
        self._systemctl(["daemon-reload"])

    def restart(self):
        log.info("Restarting {service_name}".format(service_name=self.service_name))
        self._systemctl(["restart", self.service_name])

    def stop(self):
        log.info("Stopping {service_name}".format(service_name=self.service_name))
        self._systemctl(["stop", self.service_name], stderr=subprocess.DEVNULL)

    def enable(self):
        log.info("Enabling {service_name}".format(service_name=self.service_name))
        self._systemctl(["enable", self.service_name])

    def disable(self):
        log.info("Disabling {service_name}".format(service_name=self.service_name))
        self._systemctl(["disable", self.service_name], stderr=subprocess.DEVNULL)



import logging

import portinus

from portinus import systemd
from . import checker

log = logging.getLogger(__name__)


class Service(object):

    def __init__(self, name):
        self.name = name
        self._systemd_service = systemd.Unit(name + "-monitor")
        self._systemd_timer = systemd.Unit(name + "-monitor", type="timer")
        pass

    def _generate_service_file(self):
        template = portinus.get_template("monitor.service")

        return template.render(
                name=self.name,
                )

    def _generate_timer_file(self):
        template = portinus.get_template("monitor.timer")

        return template.render(
                name=self.name,
                )

    def ensure(self):
        log.info("Creating/updating {name} monitor timer".format(name=self.name))
        self._systemd_service.ensure(content=self._generate_service_file(), restart=False, enable=False)
        self._systemd_timer.ensure(content=self._generate_timer_file())

    def remove(self):
        log.info("Removing {name} monitor timer".format(name=self.name))
        self._systemd_timer.remove()
        self._systemd_service.remove()



import subprocess
import shutil
import os
import logging

from jinja2 import Template

import portinus
from . import systemd

log = logging.getLogger(__name__)


class Service(object):

    def __init__(self, name, source=None):
        if not name:
            raise ValueError("Invalid value for 'name'")
        self.check_permissions()
        self.name = name
        self._source = ComposeSource(name, source)
        self._systemd_service = systemd.Unit(name)
        log.debug("Initialized portinus.Service for '{name}' with source: '{source}'".format(name=name, source=source))

    def check_permissions(self):
        if os.geteuid() != 0:
            log.error("You must be root to run this command!")
            raise PermissionError("You must be root to run this command!")

    def exists(self):
        return os.path.isdir(portinus.get_instance_dir(self.name))

    def _generate_service_file(self):
        start_command = "{service_script} up".format(service_script=self._source.service_script)
        stop_command = "{service_script} down".format(service_script=self._source.service_script)

        template = portinus.get_template("instance.service")
        return template.render(
            name=self.name,
            start_command=start_command,
            stop_command=stop_command,
            )

    def ensure(self):
        log.info("Creating/updating {name} portinus instance".format(name=self.name))
        try:
            self._systemd_service.stop()
        except subprocess.CalledProcessError:
            pass
        self._source.ensure()
        self._systemd_service.ensure(content=self._generate_service_file())

    def remove(self):
        log.info("Removing {name} portinus instance".format(name=self.name))
        self._systemd_service.remove()
        self._source.remove()

    def restart(self):
        log.info("Restarting {name}".format(name=self.name))
        self._systemd_service.restart()

    def compose(self, command):
        log.info("Running compose for {name} with command: '{command}'".format(name=self.name, command=command))
        if not self.exists():
            raise ValueError("The specified service does not exist")
        subprocess.call([self._source.service_script] + list(command))


class ComposeSource(object):

    def __init__(self, name, source=None):
        self.name = name
        self._source = source
        self.path = portinus.get_instance_dir(name)
        self.service_script = os.path.join(self.path, name)

        if source:
            try:
                with open(os.path.join(source, "docker-compose.yml")):
                    pass
            except Exception as e:
                log.error("Unable to access the specified source docker compose file in ({source})".format(source=source))
                raise(e)
        log.debug("Initialized ComposeSource for '{name}' from source: '{source}'".format(name=name, source=source))

    def _ensure_service_script(self):
        service_script_template = os.path.join(portinus.template_dir, "service-script")
        shutil.copy(service_script_template, self.service_script)
        os.chmod(self.service_script, 0o755)

    def ensure(self):
        if not self._source:
            log.error("No valid source specified")
            raise(IOError("No valid source specified"))
        log.info("Copying source files for '{self.name}' to '{self.path}'")
        self.remove()
        shutil.copytree(self._source, self.path, symlinks=True, copy_function=shutil.copy)
        self._ensure_service_script()
        log.debug("Successfully copied source files")

    def remove(self):
        log.info("Removing source files for '{name}' from '{path}'".format(name=self.name, path=self.path))
        try:
            shutil.rmtree(self.path)
            log.debug("Successfully removed source files")
        except FileNotFoundError:
            log.debug("No source files found")


class EnvironmentFile(object):

    def __init__(self, name, source_environment_file=None):
        self.name = name
        self._source_environment_file = source_environment_file
        self.path = portinus.get_instance_dir(self.name) + ".environment"
        log.debug("Initialized EnvironmentFile for '{name}' from source: '{source_environment_file}'".format(name=name, source_environment_file=source_environment_file))

        if source_environment_file:
            try:
                with open(source_environment_file):
                    pass
            except FileNotFoundError as e:
                log.error("Unable to access the specified environment file ({source_environment_file})".format(source_environment_file=source_environment_file))
                raise(e)

    def __bool__(self):
        return bool(self._source_environment_file)

    def ensure(self):
        if self:
            log.info("Creating/updating environment file for '{name}' at '{path}'".format(name=self.name, path=self.path))
            shutil.copy(self._source_environment_file, self.path)
        else:
            self.remove()

    def remove(self):
        log.info("Removing environment file for {name}".format(name=self.name))
        try:
            os.remove(self.path)
            log.debug("Sucessfully removed environment file")
        except FileNotFoundError:
            log.debug("No environment file found")



import logging

import portinus
from . import systemd

log = logging.getLogger(__name__)


class Timer(object):

    def __init__(self, name, restart_schedule):
        self.name = name
        self.restart_schedule = restart_schedule
        self._systemd_service = systemd.Unit(name + "-restart")
        self._systemd_timer = systemd.Unit(name + "-restart", type="timer")
        log.debug("Initialized restart.Timer for '{name}' with restart_schedule: '{restart_schedule}'".format(name=name, restart_schedule=restart_schedule))

    def __bool__(self):
        return bool(self.restart_schedule)

    def _generate_service_file(self):
        template_file = portinus.get_template("restart.service")
        instance_service = systemd.Unit(self.name)

        return template_file.render(
                name=self.name,
                service_name=instance_service.service_name,
                )

    def _generate_timer_file(self):
        template_file = portinus.get_template("restart.timer")

        return template_file.render(
                name=self.name,
                restart_schedule=self.restart_schedule,
                )

    def ensure(self):
        if self:
            log.info("Creating/updating {name} restart timer".format(name=self.name))
            self._systemd_service.ensure(content=self._generate_service_file(), restart=False, enable=False)
            self._systemd_timer.ensure(content=self._generate_timer_file())
        else:
            log.info("No restart schedule specified for {name}. Removing any existing restart timers".format(name=self.name))
            self.remove()

    def remove(self):
        log.info("Removing {name} restart timer. Errors from systemctl may occur".format(name=self.name))
        self._systemd_timer.remove()
        self._systemd_service.remove()



#!/usr/bin/env python3

import click
import logging
import sys

import portinus

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group()
@click.option('-v', '--verbose', count=True, help="Enable more logging. More -v's for more logging")
def task(verbose):
    log_level = logging.WARNING
    if verbose == 1:
        log_level = logging.INFO
    if verbose >= 2:
        log_level = logging.DEBUG
    logging.basicConfig(level=log_level)


@task.command()
@click.argument('name', required=True)
def remove(name):
    try:
        service = portinus.Application(name)
    except PermissionError:
        sys.exit(1)
    service.remove()


@task.command()
@click.argument('name', required=True)
@click.option('--source', type=click.Path(exists=True), required=True, help="A path to a folder containg a docker-compose.yml")
@click.option('--env', help="A file containing the list of environment variables to use")
@click.option('--restart', help="Provide a systemd 'OnCalender' scheduling string to force a restart of the service on the specified interval (e.g. 'weekly' or 'daily')")
def ensure(name, source, env, restart):
    try:
        service = portinus.Application(name, source=source, environment_file=env, restart_schedule=restart)
    except PermissionError:
        sys.exit(1)
    service.ensure()


@task.command()
@click.argument('name', required=True)
def restart(name)
    service = portinus.portinus.Service(name)
    service.restart()

@task.command()
@click.argument('name', required=True)
@click.argument('args', required=True, nargs=-1)
def compose(name, args):
    try:
        service = portinus.portinus.Service(name)
        service.compose(args)
    except (PermissionError, ValueError) as e:
        print(e)
        sys.exit(1)


if __name__ == "__main__":
    task()
